@@ -6,7 +6,7 @@ use Carp ();
 use IO::Handle ();
 use Exporter 'import';
 
-our $VERSION = '0.002';
+our $VERSION = '0.003';
 
 our @EXPORT = 'cgi';
 
@@ -78,41 +78,87 @@ my %HTTP_STATUS = (
     511 => 'Network Authentication Required', # RFC 6585: Additional Codes
 );
 
+my @DAYS_OF_WEEK = qw(Sun Mon Tue Wed Thu Fri Sat);
+my @MONTH_NAMES = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
+my %MONTH_NUMS;
+@MONTH_NUMS{@MONTH_NAMES} = 0..11;
+
+sub epoch_to_date {
+  my ($sec,$min,$hour,$mday,$mon,$year,$wday) = gmtime $_[0];
+  return sprintf '%s, %02d %s %04d %02d:%02d:%02d GMT',
+    $DAYS_OF_WEEK[$wday], $mday, $MONTH_NAMES[$mon], $year + 1900, $hour, $min, $sec;
+}
+
+sub date_to_epoch {
+  # RFC 1123 (Sun, 06 Nov 1994 08:49:37 GMT)
+  my ($mday,$mon,$year,$hour,$min,$sec) = $_[0] =~ m/^ (?:Sun|Mon|Tue|Wed|Thu|Fri|Sat),
+    [ ] ([0-9]{2}) [ ] (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) [ ] ([0-9]{4})
+    [ ] ([0-9]{2}) : ([0-9]{2}) : ([0-9]{2}) [ ] GMT $/x;
+
+  # RFC 850 (Sunday, 06-Nov-94 08:49:37 GMT)
+  ($mday,$mon,$year,$hour,$min,$sec) = $_[0] =~ m/^ (?:Sun|Mon|Tues|Wednes|Thurs|Fri|Satur)day,
+    [ ] ([0-9]{2}) - (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) - ([0-9]{2})
+    [ ] ([0-9]{2}) : ([0-9]{2}) : ([0-9]{2}) [ ] GMT $/x unless defined $mday;
+
+  # asctime (Sun Nov  6 08:49:37 1994)
+  ($mon,$mday,$hour,$min,$sec,$year) = $_[0] =~ m/^ (?:Sun|Mon|Tue|Wed|Thu|Fri|Sat)
+    [ ] (Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec) [ ]{1,2} ([0-9]{1,2})
+    [ ] ([0-9]{2}) : ([0-9]{2}) : ([0-9]{2}) [ ] ([0-9]{4}) $/x unless defined $mday;
+
+  return undef unless defined $mday;
+
+  require Time::Local;
+  # 4 digit years interpreted literally, but may have leading zeroes
+  # 2 digit years interpreted with best effort heuristic
+  return scalar Time::Local::timegm($sec, $min, $hour, $mday, $MONTH_NUMS{$mon},
+    (length($year) == 4 && $year < 1900) ? $year - 1900 : $year);
+}
+
+# for cleanup in END in case of premature exit
+my %PENDING_CGI;
+
 sub cgi (&) {
   my ($handler) = @_;
   my $cgi = bless {}, __PACKAGE__;
+  $PENDING_CGI{0+$cgi} = $cgi;
   my ($error, $errored);
   {
     local $@;
-    eval { $handler->($cgi); die "No response rendered by cgi\n" unless $cgi->{headers_rendered}; 1 }
+    eval { local $_ = $cgi; $handler->(); die "cgi completed without rendering a response\n" unless $cgi->{headers_rendered}; 1 }
       or do { $error = $@; $errored = 1 };
   }
-  if ($errored) {
-    my $caller = caller;
-    $cgi->set_response_status(500) unless $cgi->{headers_rendered} or defined $cgi->{response_status};
-    if (defined(my $handler = $cgi->{on_error})) {
-      my ($error_error, $error_errored);
-      {
-        local $@;
-        eval { $handler->($cgi, $error); 1 } or do { $error_error = $@; $error_errored = 1 };
-      }
-      if ($error_errored) {
-        warn "Exception in error handler: $error_error";
-        warn "Original error: $error";
-      }
-    } else {
-      warn $error;
-    }
-    $cgi->render(text => 'Internal Server Error') unless $cgi->{headers_rendered};
-  }
+  _handle_error($cgi, $error) if $errored;
+  delete $PENDING_CGI{0+$cgi};
   1;
 }
 
-sub set_error_handler { $_[0]{on_error} = $_[1]; $_[0] }
-sub request_body_limit {
-  return defined $_[0]{request_body_limit} ? $_[0]{request_body_limit} :
-    ($_[0]{request_body_limit} = defined $ENV{CGI_TINY_REQUEST_BODY_LIMIT} ? $ENV{CGI_TINY_REQUEST_BODY_LIMIT} : 16777216);
+END {
+  foreach my $key (keys %PENDING_CGI) {
+    my $cgi = delete $PENDING_CGI{$key};
+    _handle_error($cgi, "cgi exited without rendering a response\n") unless $cgi->{headers_rendered};
+  }
 }
+
+sub _handle_error {
+  my ($cgi, $error) = @_;
+  $cgi->{response_status} = "500 $HTTP_STATUS{500}" unless $cgi->{headers_rendered} or defined $cgi->{response_status};
+  if (defined(my $handler = $cgi->{on_error})) {
+    my ($error_error, $error_errored);
+    {
+      local $@;
+      eval { $handler->($cgi, $error); 1 } or do { $error_error = $@; $error_errored = 1 };
+    }
+    if ($error_errored) {
+      warn "Exception in error handler: $error_error";
+      warn "Original error: $error";
+    }
+  } else {
+    warn $error;
+  }
+  $cgi->render(text => 'Internal Server Error') unless $cgi->{headers_rendered};
+}
+
+sub set_error_handler { $_[0]{on_error} = $_[1]; $_[0] }
 sub set_request_body_limit { $_[0]{request_body_limit} = $_[1]; $_[0] }
 sub set_input_handle       { $_[0]{input_handle} = $_[1]; $_[0] }
 sub set_output_handle      { $_[0]{output_handle} = $_[1]; $_[0] }
@@ -186,13 +232,33 @@ sub _headers {
   return $self->{request_headers};
 }
 
+sub cookies { +{%{$_[0]->_cookies}} }
+sub cookie { $_[0]->_cookies->{$_[1]} }
+
+sub _cookies {
+  my ($self) = @_;
+  unless (exists $self->{request_cookies}) {
+    $self->{request_cookies} = {};
+    if (defined $ENV{HTTP_COOKIE}) {
+      foreach my $pair (split /\s*;\s*/, $ENV{HTTP_COOKIE}) {
+        next unless length $pair;
+        my ($name, $value) = split /=/, $pair, 2;
+        $self->{request_cookies}{$name} = $value if defined $value;
+      }
+    }
+  }
+  return $self->{request_cookies};
+}
+
 sub body {
   my ($self) = @_;
   unless (exists $self->{content}) {
-    my $limit = $self->request_body_limit;
+    my $limit = $self->{request_body_limit};
+    $limit = $ENV{CGI_TINY_REQUEST_BODY_LIMIT} unless defined $limit;
+    $limit = 16777216 unless defined $limit;
     my $length = $ENV{CONTENT_LENGTH} || 0;
     if ($limit and $length > $limit) {
-      $self->set_response_status(413);
+      $self->{response_status} = "413 $HTTP_STATUS{413}" unless $self->{headers_rendered};
       die "Request body limit exceeded\n";
     }
     $_[0]{content} = '';
@@ -228,8 +294,7 @@ sub _body_params {
   my ($self) = @_;
   unless (exists $self->{body_params}) {
     my (@ordered, %keyed);
-    my $content_type = $self->content_type;
-    if (defined $content_type and $content_type =~ m/^application\/x-www-form-urlencoded/i) {
+    if ($ENV{CONTENT_TYPE} and $ENV{CONTENT_TYPE} =~ m/^application\/x-www-form-urlencoded/i) {
       foreach my $pair (split /&/, $self->body) {
         my ($key, $value) = split /=/, $pair, 2;
         $value = '' unless defined $value;
@@ -246,8 +311,7 @@ sub _body_params {
 sub body_json {
   my ($self) = @_;
   unless (exists $self->{body_json}) {
-    my $content_type = $self->content_type;
-    if (defined $content_type and $content_type =~ m/^application\/json/i) {
+    if ($ENV{CONTENT_TYPE} and $ENV{CONTENT_TYPE} =~ m/^application\/json/i) {
       $self->{body_json} = $self->_json->decode($self->body);
     }
   }
@@ -285,9 +349,45 @@ sub add_response_header {
   return $self;
 }
 
+my %COOKIE_ATTR_VALUE = (expires => 1, domain => 1, path => 1, secure => 0, httponly => 0, samesite => 1, 'max-age' => 1);
+sub add_response_cookie {
+  my ($self, $name, $value, @attrs) = @_;
+  if ($self->{headers_rendered}) {
+    Carp::carp "Attempted to add HTTP response cookie '$name' but headers have already been rendered";
+  } else {
+    my $cookie_str = "$name=$value";
+    my $i = 0;
+    while ($i <= $#attrs) {
+      my ($key, $val) = @attrs[$i, $i+1];
+      my $has_value = $COOKIE_ATTR_VALUE{lc $key};
+      if (!defined $has_value) {
+        Carp::carp "Attempted to set unknown cookie attribute '$key'";
+      } elsif ($has_value) {
+        $cookie_str .= "; $key=$val" if defined $val;
+      } else {
+        $cookie_str .= "; $key" if $val;
+      }
+    } continue {
+      $i += 2;
+    }
+    push @{$self->{response_headers}}, ['Set-Cookie', $cookie_str];
+  }
+  return $self;
+}
+
 sub response_charset     { defined $_[0]{response_charset} ? $_[0]{response_charset} : ($_[0]{response_charset} = 'UTF-8') }
 sub set_response_charset { $_[0]{response_charset} = $_[1]; $_[0] }
 
+sub set_nph {
+  my ($self, $value) = @_;
+  if ($self->{headers_rendered}) {
+    Carp::carp "Attempted to set NPH response mode but headers have already been rendered";
+  } else {
+    $self->{nph} = $value;
+  }
+  return $self;
+}
+
 sub headers_rendered { $_[0]{headers_rendered} }
 
 my %known_types = (json => 1, html => 1, xml => 1, text => 1, data => 1, redirect => 1);
@@ -321,16 +421,24 @@ sub render {
         unless defined $content_type;
       $headers_str = "Content-Type: $content_type\r\n$headers_str";
     }
-    if (!$headers_set{status}) {
-      if (defined(my $status = $self->{response_status})) {
-        $headers_str = "Status: $status\r\n$headers_str";
-      } elsif ($type eq 'redirect') {
-        $headers_str = "Status: 302 $HTTP_STATUS{302}\r\n$headers_str";
-      }
+    if (!$headers_set{date}) {
+      my $date_str = epoch_to_date(time);
+      $headers_str = "Date: $date_str\r\n$headers_str";
+    }
+    my $status = $self->{response_status};
+    $status = "302 $HTTP_STATUS{302}" if !defined $status and $type eq 'redirect';
+    if ($self->{nph}) {
+      $status = "200 $HTTP_STATUS{200}" unless defined $status;
+      my $protocol = $ENV{SERVER_PROTOCOL};
+      $protocol = 'HTTP/1.0' unless defined $protocol and length $protocol;
+      $headers_str = "$protocol $status\r\n$headers_str";
+      my $server = $ENV{SERVER_SOFTWARE};
+      $headers_str .= "Server: $server\r\n" if defined $server and length $server;
+    } elsif (!$headers_set{status} and defined $status) {
+      $headers_str = "Status: $status\r\n$headers_str";
     }
-    $headers_str .= "\r\n";
     binmode $out_fh;
-    $out_fh->printflush($headers_str);
+    $out_fh->printflush("$headers_str\r\n");
     $self->{headers_rendered} = 1;
   } elsif ($type eq 'redirect') {
     Carp::carp "Attempted to render a redirect but headers have already been rendered";
@@ -374,7 +482,7 @@ CGI::Tiny - Common Gateway Interface, with no frills
   use CGI::Tiny;
 
   cgi {
-    my ($cgi) = @_;
+    my $cgi = $_;
     $cgi->set_error_handler(sub {
       my ($cgi, $error) = @_;
       warn $error;
@@ -388,7 +496,8 @@ CGI::Tiny - Common Gateway Interface, with no frills
       $fribble = $cgi->body_param('fribble');
     } else {
       $cgi->set_response_status(405);
-      return $cgi->render;
+      $cgi->render;
+      exit;
     }
     die "Invalid fribble parameter" unless length $fribble;
     $cgi->render(json => {fribble => $fribble});
@@ -436,6 +545,8 @@ rendering output.
 
 =back
 
+See L</"COMPARISON TO CGI.PM">.
+
 This module's interface is currently I<EXPERIMENTAL> and may be changed
 incompatibly if needed.
 
@@ -446,14 +557,14 @@ incompatibly if needed.
 CGI::Tiny's interface is a regular function called C<cgi> exported by default.
 
   cgi {
-    my ($cgi) = @_;
+    my $cgi = $_;
     # set up error handling on $cgi
     # inspect request data via $cgi
     # set response headers if needed via $cgi
     # render response data with $cgi->render
   };
 
-The code block is immediately run and passed a CGI::Tiny object, which
+The code block is immediately run with C<$_> set to a CGI::Tiny object, which
 L</"METHODS"> can be called on to read request information and render a
 response.
 
@@ -462,6 +573,12 @@ render a response, it will run the handler set by L</"set_error_handler"> if
 any, or by default emit the error as a warning and (if nothing has been
 rendered yet) render a 500 Internal Server Error.
 
+Note that the C<cgi> block's current implementation as a regular exported
+subroutine is an implementation detail, and future implementations reserve the
+right to provide it as an XSUB or keyword for performance reasons. You should
+not rely on C<@_> to be set, and you should not use C<return> to exit the
+block; use C<exit> to end a CGI script early after rendering a response.
+
 =head1 EXTENDING
 
 CGI::Tiny is a minimal interface to the CGI protocol, but can be extended with
@@ -490,7 +607,7 @@ L<Text::Xslate> is an efficient template engine designed for HTML/XML.
   use Data::Section::Simple 'get_data_section';
 
   cgi {
-    my ($cgi) = @_;
+    my $cgi = $_;
     my $foo = $cgi->query_param('foo');
     my $tx = Text::Xslate->new(path => ['templates'], cache => 0);
 
@@ -519,7 +636,7 @@ toolkit.
   use Mojo::Loader 'data_section';
 
   cgi {
-    my ($cgi) = @_;
+    my $cgi = $_;
     my $foo = $cgi->query_param('foo');
     my $mt = Mojo::Template->new(auto_escape => 1, vars => 1);
 
@@ -568,7 +685,7 @@ name).
   );
 
   cgi {
-    my ($cgi) = @_;
+    my $cgi = $_;
 
     my $routes = Routes::Tiny->new;
     # /script.cgi/foo
@@ -612,23 +729,17 @@ error handlers should render some response if L</"headers_rendered"> is
 false. If no response has been rendered after the error handler completes, the
 default 500 Internal Server Error response will be rendered.
 
-=head3 request_body_limit
-
-  my $limit = $cgi->request_body_limit;
-
-Limit in bytes for parsing a request body into memory, defaults to the value of
-the C<CGI_TINY_REQUEST_BODY_LIMIT> environment variable or 16777216 (16 MiB).
-Since the request body is not parsed until needed, methods that parse the whole
-request body into memory like L</"body"> will set the response status to
-C<413 Payload Too Large> and throw an exception if the content length is over
-the limit. A value of 0 will remove the limit (not recommended unless you have
-other safeguards on memory usage).
-
 =head3 set_request_body_limit
 
   $cgi = $cgi->set_request_body_limit(16*1024*1024);
 
-Sets L</"request_body_limit">.
+Sets the limit in bytes for parsing a request body into memory. If not set,
+defaults to the value of the C<CGI_TINY_REQUEST_BODY_LIMIT> environment
+variable or 16777216 (16 MiB). Since the request body is not parsed until
+needed, methods that parse the whole request body into memory like L</"body">
+will set the response status to C<413 Payload Too Large> and throw an exception
+if the content length is over the limit. A value of 0 will remove the limit
+(not recommended unless you have other safeguards on memory usage).
 
 =head3 set_input_handle
 
@@ -745,6 +856,18 @@ Retrieve the value of a request header by name (case insensitive). CGI request
 headers can only contain a single value, which may be combined from multiple
 values.
 
+=head3 cookies
+
+  my $hashref = $cgi->cookies;
+
+Hash reference of request cookie names and values.
+
+=head3 cookie
+
+  my $value = $cgi->cookie('foo');
+
+Retrieve the value of a request cookie by name.
+
 =head3 body
 
   my $bytes = $cgi->body;
@@ -752,7 +875,7 @@ values.
 Retrieve the request body as bytes.
 
 Note that this will read the whole request body into memory, so make sure the
-L</"request_body_limit"> can fit well within the available memory.
+L</"set_request_body_limit"> can fit well within the available memory.
 
 =head3 body_pairs
 
@@ -762,7 +885,7 @@ Retrieve C<x-www-form-urlencoded> body parameters as an array reference of
 two-element array references.
 
 Note that this will read the whole request body into memory, so make sure the
-L</"request_body_limit"> can fit well within the available memory.
+L</"set_request_body_limit"> can fit well within the available memory.
 
 =head3 body_params
 
@@ -772,7 +895,7 @@ Retrieve C<x-www-form-urlencoded> body parameters as a hash reference. If a
 parameter name is passed multiple times, its value will be an array reference.
 
 Note that this will read the whole request body into memory, so make sure the
-L</"request_body_limit"> can fit well within the available memory.
+L</"set_request_body_limit"> can fit well within the available memory.
 
 =head3 body_param
 
@@ -783,7 +906,7 @@ parameter name is passed multiple times, returns the last value. Use
 L</"body_param_array"> to get multiple values of a parameter.
 
 Note that this will read the whole request body into memory, so make sure the
-L</"request_body_limit"> can fit well within the available memory.
+L</"set_request_body_limit"> can fit well within the available memory.
 
 =head3 body_param_array
 
@@ -793,7 +916,7 @@ Retrieve values of a named C<x-www-form-urlencoded> body parameter as an array
 reference.
 
 Note that this will read the whole request body into memory, so make sure the
-L</"request_body_limit"> can fit well within the available memory.
+L</"set_request_body_limit"> can fit well within the available memory.
 
 =head3 body_json
 
@@ -802,7 +925,7 @@ L</"request_body_limit"> can fit well within the available memory.
 Decode an C<application/json> request body from UTF-8-encoded JSON.
 
 Note that this will read the whole request body into memory, so make sure the
-L</"request_body_limit"> can fit well within the available memory.
+L</"set_request_body_limit"> can fit well within the available memory.
 
 =head2 Response
 
@@ -832,6 +955,82 @@ deduplicate or munge headers that have been added manually. Headers are printed
 in the response in the same order added, and adding the same header multiple
 times will result in multiple instances of that response header.
 
+=head3 add_response_cookie
+
+  $cgi = $cgi->add_response_cookie($name => $value,
+    Expires   => 'Sun, 06 Nov 1994 08:49:37 GMT',
+    HttpOnly  => 1,
+    'Max-Age' => 3600,
+    Path      => '/foo',
+    SameSite  => 'Strict',
+    Secure    => 1,
+  );
+
+Adds a response cookie. No effect after response headers have been rendered.
+
+Note that cookie values should only consist of ASCII characters and may not
+contain any control characters, space characters, or the characters C<",;\>.
+More complex values can be encoded to UTF-8 and L<base64|MIME::Base64> for
+transport.
+
+  use Encode 'encode';
+  use MIME::Base64 'encode_base64';
+  $cgi->add_response_cookie(foo => encode_base64(encode('UTF-8', $value), ''));
+
+  use Encode 'decode';
+  use MIME::Base64 'decode_base64';
+  my $value = decode 'UTF-8', decode_base64 $cgi->cookie('foo');
+
+Structures can be encoded to JSON and base64 for transport.
+
+  use Cpanel::JSON::XS 'encode_json';
+  use MIME::Base64 'encode_base64';
+  $cgi->add_response_cookie(foo => encode_base64(encode_json(\%hash), ''));
+
+  use Cpanel::JSON::XS 'decode_json';
+  use MIME::Base64 'decode_base64';
+  my $hashref = decode_json decode_base64 $cgi->cookie('foo');
+
+Optional cookie attributes are specified in key-value pairs after the cookie
+name and value. Cookie attribute names are case-insensitive.
+
+=over
+
+=item Domain
+
+Domain for which cookie is valid.
+
+=item Expires
+
+Expiration date string for cookie. L</"epoch_to_date"> can be used to generate
+the appropriate date string format.
+
+=item HttpOnly
+
+If set to a true value, the cookie will be restricted from client-side scripts.
+
+=item Max-Age
+
+Max age of cookie before it expires, in seconds, as an alternative to
+specifying C<Expires>.
+
+=item Path
+
+URL path for which cookie is valid.
+
+=item SameSite
+
+C<Strict> to restrict the cookie to requests from the same site, C<Lax> to
+allow it additionally in certain cross-site requests. This attribute is
+currently part of a draft specification so its handling may change, but it is
+supported by most browsers.
+
+=item Secure
+
+If set to a true value, the cookie will be restricted to HTTPS requests.
+
+=back
+
 =head3 response_charset
 
   my $charset = $cgi->response_charset;
@@ -845,6 +1044,18 @@ defaults to C<UTF-8>.
 
 Sets L</"response_charset">.
 
+=head3 set_nph
+
+  $cgi = $cgi->set_nph(1);
+
+If set to a true value before rendering response headers, CGI::Tiny will act as
+a L<NPH (Non-Parsed Header)|https://tools.ietf.org/html/rfc3875#section-5>
+script and render full HTTP response headers. This may be required for some CGI
+servers, or enable unbuffered responses or HTTP extensions not supported by the
+CGI server.
+
+No effect after response headers have been rendered.
+
 =head3 headers_rendered
 
   my $bool = $cgi->headers_rendered;
@@ -876,11 +1087,43 @@ C<html>, C<xml>, or C<text> data is expected to be decoded characters, and will
 be encoded according to L</"response_charset">. C<json> data will be encoded to
 UTF-8.
 
-C<redirect> will print a redirection header if response headers have not yet
-been rendered, and will set a response status of 302 if none has been set by
+C<redirect> will set a L<Location> header if response headers have not yet been
+rendered, and will set a response status of 302 if none has been set by
 L</"set_response_status">. It will not set a C<Content-Type> response header.
 If response headers have already been rendered a warning will be emitted.
 
+The C<Date> response header will be set to the current time as an HTTP date
+string if not set manually.
+
+=head1 FUNCTIONS
+
+The following convenience functions are provided but not exported.
+
+=head2 epoch_to_date
+
+  my $date = CGI::Tiny::epoch_to_date $epoch;
+
+Convert a Unix epoch timestamp to a RFC 1123 HTTP date string suitable for use
+in HTTP headers such as C<Date> and C<Expires>.
+
+=head2 date_to_epoch
+
+  my $epoch = CGI::Tiny::date_to_epoch $date;
+
+Parse a RFC 1123 HTTP date string to a Unix epoch timestamp. For compatibility
+as required by L<RFC 7231|https://tools.ietf.org/html/rfc7231#section-7.1.1.1>,
+legacy RFC 850 and ANSI C asctime date formats are also recognized. Returns
+C<undef> if the string does not parse as any of these formats.
+
+  # RFC 1123
+  my $epoch = CGI::Tiny::date_to_epoch 'Sun, 06 Nov 1994 08:49:37 GMT';
+
+  # RFC 850
+  my $epoch = CGI::Tiny::date_to_epoch 'Sunday, 06-Nov-94 08:49:37 GMT';
+
+  # asctime
+  my $epoch = CGI::Tiny::date_to_epoch 'Sun Nov  6 08:49:37 1994';
+
 =head1 ENVIRONMENT
 
 CGI::Tiny recognizes the following environment variables, in addition to the
@@ -888,7 +1131,84 @@ standard CGI environment variables.
 
 =head2 CGI_TINY_REQUEST_BODY_LIMIT
 
-Default value for L</"request_body_limit">.
+Default value for L</"set_request_body_limit">.
+
+=head1 COMPARISON TO CGI.PM
+
+Traditionally, the L<CGI> module (referred to as CGI.pm to differentiate it
+from the CGI protocol) has been used to write Perl CGI scripts. This module
+fills a similar need but has a number of interface differences to be aware of.
+
+=over
+
+=item *
+
+There is no global CGI::Tiny object; the object is constructed for the scope of
+the C<cgi> block, only reads request data from the environment once it is
+accessed, and once the block completes (normally or abnormally), it ensures
+that a valid response is rendered to avoid gateway errors.
+
+=item *
+
+Instead of global variables like C<$CGI::POST_MAX>, global behavior settings
+are applied to the CGI::Tiny object inside the C<cgi> block.
+
+=item *
+
+Exceptions within the C<cgi> block are handled by default by rendering a server
+error response and emitting the error as a warning. This can be customized with
+L</"set_error_handler">.
+
+=item *
+
+Request query and body parameter accessors in CGI::Tiny are not context
+sensitive. L</"query_param"> and L</"body_param"> always return a single value,
+and L</"query_param_array"> and L</"body_param_array"> must be used to retrieve
+multi-value parameters. CGI::Tiny also does not have a method-sensitive
+C<param> accessor; query or body parameters must be accessed specifically.
+
+=item *
+
+CGI::Tiny decodes request query and body parameters from UTF-8 to Unicode
+characters by default, and L</"render"> provides methods to encode response
+data from Unicode characters to UTF-8 or other charsets automatically.
+
+=item *
+
+In CGI.pm, response headers must be printed manually before any response data
+is printed to avoid malformed responses. In CGI::Tiny, the L</"render"> method
+is used to print response data, and automatically prints response headers the
+first time it is called. C<redirect> responses are also handled by
+L</"render">.
+
+=item *
+
+In CGI::Tiny, a custom response status is set by calling
+L</"set_response_status"> before the first L</"render">, which only requires
+the status code and will add the appropriate human-readable status message
+itself.
+
+=item *
+
+Response setters are distinct methods from request accessors in CGI::Tiny.
+L</"content_type">, L</"header">, and L</"cookie"> are used to access request
+data, and L</"set_response_content_type">, L</"add_response_header">, and
+L</"add_response_cookie"> are used to set response headers for the pending
+response before the first call to L</"render">.
+
+=item *
+
+CGI::Tiny does not provide any HTML templating helpers, as this functionality
+is much better implemented by other robust implementations on CPAN; see
+L</"Templating">.
+
+=item *
+
+CGI::Tiny does not do any implicit encoding of cookie values or the C<Expires>
+header or cookie attribute. The L</"epoch_to_date"> convenience function is
+provided to render appropriate C<Expires> date values.
+
+=back
 
 =head1 CAVEATS
 
@@ -897,6 +1217,10 @@ state of environment variables and the C<STDIN> and C<STDOUT> standard
 filehandles. CGI::Tiny does not prevent you from messing with these interfaces
 directly, but it may result in confusion.
 
+CGI::Tiny eschews certain sanity checking for performance reasons. For example,
+Content-Type and other header values set for the response should only contain
+ASCII text with no control characters, but CGI::Tiny does not verify this.
+
 Most applications are better written in a L<PSGI>-compatible framework (e.g.
 L<Dancer2> or L<Mojolicious>) and deployed in a persistent application server
 so that the application does not have to start up again every time it receives
@@ -908,10 +1232,6 @@ a request.
 
 =item * Uploads/multipart request
 
-=item * Cookies
-
-=item * NPH
-
 =back
 
 =head1 BUGS
